import copy
import gc
import logging
import os
import random
import sys
import unittest

import pyvex
from pyvex.lifting import LibVEXLifter

if sys.platform == "linux":
    import resource


# pylint: disable=R0201
class TestPyvex(unittest.TestCase):
    @unittest.skipUnless(
        sys.platform == "linux", "Cannot import the resource package on windows, values different on macos."
    )
    def test_memory(self):
        arches = [pyvex.ARCH_X86, pyvex.ARCH_PPC32, pyvex.ARCH_AMD64, pyvex.ARCH_ARM_BE]
        # we're not including ArchMIPS32 cause it segfaults sometimes

        # disable logging, as that may fill up log buffers somewhere
        logging.disable(logging.ERROR)

        for _ in range(10000):
            try:
                s = os.urandom(32)
                a = random.choice(arches)
                p = pyvex.IRSB(data=s, mem_addr=0, arch=a)
            except pyvex.PyVEXError:
                pass

        kb_start = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

        for _ in range(20000):
            try:
                s = os.urandom(32)
                a = random.choice(arches)
                p = pyvex.IRSB(data=s, mem_addr=0, arch=a)
            except pyvex.PyVEXError:
                pass
        del p
        gc.collect()

        logging.disable(logging.NOTSET)

        kb_end = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

        pyvex.pvc.clear_log()
        pyvex.pvc.LibVEX_ShowAllocStats()
        print(LibVEXLifter.get_vex_log())

        # allow a 5mb leeway
        assert kb_end - kb_start < 5000

    ################
    ### IRCallee ###
    ################

    def test_ircallee(self):
        callee = pyvex.IRCallee(3, "test_name", 0xFFFFFF)
        assert callee.name == "test_name"
        assert callee.regparms == 3
        assert callee.mcx_mask == 0xFFFFFF

    ############
    ### IRSB ###
    ############

    def test_irsb_empty(self):
        self.assertRaises(Exception, pyvex.IRSB)
        self.assertRaises(Exception, pyvex.IRSB, data="", arch=pyvex.ARCH_AMD64, mem_addr=0)

    def test_irsb_arm(self):
        irsb = pyvex.IRSB(data=b"\x33\xff\x2f\xe1", mem_addr=0, arch=pyvex.ARCH_ARM_BE)
        assert len([i for i in irsb.statements if isinstance(i, pyvex.IRStmt.IMark)]) == 1

    def test_irsb_popret(self):
        irsb = pyvex.IRSB(data=b"\x5d\xc3", mem_addr=0, arch=pyvex.ARCH_AMD64)
        stmts = irsb.statements
        irsb.pp()

        assert len(stmts) > 0
        assert irsb.jumpkind == "Ijk_Ret"
        assert irsb.offsIP == 184

        cursize = len(irsb.tyenv.types)
        assert cursize > 0
        print(irsb.statements[10].data)
        print(irsb.statements[10].data.tmp)
        print(irsb.tyenv.types[irsb.statements[10].data.tmp])
        assert irsb.tyenv.lookup(irsb.statements[10].data.tmp) == "Ity_I64"

    def test_two_irsb(self):
        irsb1 = pyvex.IRSB(data=b"\x5d\xc3", mem_addr=0, arch=pyvex.ARCH_AMD64)
        irsb2 = pyvex.IRSB(data=b"\x5d\x5d\x5d\x5d", mem_addr=0, arch=pyvex.ARCH_AMD64)

        stmts1 = irsb1.statements
        stmts2 = irsb2.statements

        assert len(stmts1) != len(stmts2)

    def test_irsb_deepCopy(self):
        irsb = pyvex.IRSB(data=b"\x5d\xc3", mem_addr=0, arch=pyvex.ARCH_AMD64)
        stmts = irsb.statements

        irsb2 = copy.deepcopy(irsb)
        stmts2 = irsb2.statements
        assert len(stmts) == len(stmts2)

    def test_irsb_addStmt(self):
        irsb = pyvex.IRSB(data=b"\x5d\xc3", mem_addr=0, arch=pyvex.ARCH_AMD64)
        stmts = irsb.statements

        irsb2 = copy.deepcopy(irsb)
        irsb2.statements = []
        assert len(irsb2.statements) == 0

        for n, i in enumerate(stmts):
            assert len(irsb2.statements) == n
            irsb2.statements.append(copy.deepcopy(i))

        irsb2.pp()

    def test_irsb_tyenv(self):
        irsb = pyvex.IRSB(data=b"\x5d\xc3", mem_addr=0, arch=pyvex.ARCH_AMD64)
        print(irsb.tyenv)
        print("Orig")
        print(irsb.tyenv)

        print("Empty")
        irsb2 = pyvex.IRSB.empty_block(arch=pyvex.ARCH_AMD64, addr=0)
        print(irsb2.tyenv)

        print("Unwrapped")
        irsb2.tyenv = copy.deepcopy(irsb.tyenv)
        print(irsb2.tyenv)

    ##################
    ### Statements ###
    ##################

    def test_irstmt_pp(self):
        irsb = pyvex.IRSB(data=b"\x5d\xc3", mem_addr=0, arch=pyvex.ARCH_AMD64)
        stmts = irsb.statements
        for i in stmts:
            print("STMT: ", end=" ")
            print(i)

    def test_irstmt_flat(self):
        print("TODO")

    def test_irstmt_imark(self):
        m = pyvex.IRStmt.IMark(1, 2, 3)
        assert m.tag == "Ist_IMark"
        assert m.addr == 1
        assert m.len == 2
        assert m.delta == 3

        m.addr = 5
        assert m.addr == 5
        m.len = 5
        assert m.len == 5
        m.delta = 5
        assert m.delta == 5

        self.assertRaises(Exception, pyvex.IRStmt.IMark, ())

    def test_irstmt_abihint(self):
        self.assertRaises(Exception, pyvex.IRStmt.AbiHint, ())

        a = pyvex.IRExpr.RdTmp.get_instance(123)
        b = pyvex.IRExpr.RdTmp.get_instance(456)

        m = pyvex.IRStmt.AbiHint(a, 10, b)
        assert m.base.tmp == 123
        assert m.len == 10
        assert m.nia.tmp == 456

    def test_irstmt_put(self):
        self.assertRaises(Exception, pyvex.IRStmt.Put, ())

        a = pyvex.IRExpr.RdTmp.get_instance(123)
        m = pyvex.IRStmt.Put(a, 10)
        print("Put stmt:", end=" ")
        print(m)
        print("")
        assert m.data.tmp == 123
        assert m.offset == 10

    def test_irexpr_puti(self):
        r = pyvex.IRRegArray(10, "Ity_I64", 20)
        i = pyvex.IRExpr.RdTmp.get_instance(5)
        d = pyvex.IRExpr.RdTmp.get_instance(30)
        m = pyvex.IRStmt.PutI(r, i, d, 2)
        assert m.descr.base == 10
        assert m.ix.tmp == 5
        assert m.bias == 2
        assert m.data.tmp == d.tmp

        self.assertRaises(Exception, pyvex.IRStmt.PutI, ())

    def test_irstmt_wrtmp(self):
        self.assertRaises(Exception, pyvex.IRStmt.WrTmp, ())

        a = pyvex.IRExpr.RdTmp.get_instance(123)
        m = pyvex.IRStmt.WrTmp(10, a)
        assert m.tag == "Ist_WrTmp"
        assert m.tmp == 10
        assert m.data.tmp == 123

    def test_irstmt_store(self):
        self.assertRaises(Exception, pyvex.IRStmt.Store, ())

        a = pyvex.IRExpr.RdTmp.get_instance(123)
        d = pyvex.IRExpr.RdTmp.get_instance(456)
        m = pyvex.IRStmt.Store(a, d, "Iend_LE")
        assert m.tag == "Ist_Store"
        assert m.endness == "Iend_LE"
        assert m.addr.tmp == a.tmp
        assert m.data.tmp == d.tmp

    def test_irstmt_cas(self):
        self.assertRaises(Exception, pyvex.IRStmt.CAS, ())

        a = pyvex.IRExpr.RdTmp.get_instance(10)
        eh = pyvex.IRExpr.RdTmp.get_instance(11)
        el = pyvex.IRExpr.RdTmp.get_instance(12)
        dh = pyvex.IRExpr.RdTmp.get_instance(21)
        dl = pyvex.IRExpr.RdTmp.get_instance(22)

        args = {
            "oldHi": 1,
            "oldLo": 2,
            "end": "Iend_LE",
            "addr": a,
            "expdHi": eh,
            "expdLo": el,
            "dataHi": dh,
            "dataLo": dl,
        }

        m = pyvex.IRStmt.CAS(**args)
        assert m.tag == "Ist_CAS"
        assert m.endness == "Iend_LE"
        assert m.oldHi == 1
        assert m.oldLo == 2
        assert m.addr.tmp == a.tmp
        assert m.expdHi.tmp == eh.tmp
        assert m.expdLo.tmp == el.tmp
        assert m.dataHi.tmp == dh.tmp
        assert m.dataLo.tmp == dl.tmp

    def test_irstmt_loadg(self):
        self.assertRaises(Exception, pyvex.IRStmt.LoadG, ())

        a = pyvex.IRExpr.RdTmp.get_instance(10)
        alt = pyvex.IRExpr.RdTmp.get_instance(11)
        guard = pyvex.IRExpr.RdTmp.get_instance(12)

        args = {
            "dst": 1,
            "end": "Iend_LE",
            "addr": a,
            "alt": alt,
            "guard": guard,
            "cvt": "ILGop_Ident32",
        }

        m = pyvex.IRStmt.LoadG(**args)
        assert m.tag == "Ist_LoadG"
        assert m.end == "Iend_LE"
        assert m.cvt == "ILGop_Ident32"
        assert m.dst == 1
        assert m.addr.tmp == a.tmp
        assert m.alt.tmp == alt.tmp
        assert m.guard.tmp == guard.tmp

        assert m.cvt_types == ("Ity_I32", "Ity_I32")

    def test_irstmt_storeg(self):
        self.assertRaises(Exception, pyvex.IRStmt.LoadG, ())

        a = pyvex.IRExpr.RdTmp.get_instance(10)
        data = pyvex.IRExpr.RdTmp.get_instance(11)
        guard = pyvex.IRExpr.RdTmp.get_instance(12)

        args = {"end": "Iend_LE", "addr": a, "data": data, "guard": guard}

        m = pyvex.IRStmt.StoreG(**args)
        assert m.tag == "Ist_StoreG"
        assert m.end == "Iend_LE"
        assert m.addr.tmp == a.tmp
        assert m.data.tmp == data.tmp
        assert m.guard.tmp == guard.tmp

    def test_irstmt_llsc(self):
        self.assertRaises(Exception, pyvex.IRStmt.LLSC)

        a = pyvex.IRExpr.RdTmp.get_instance(123)
        d = pyvex.IRExpr.RdTmp.get_instance(456)
        m = pyvex.IRStmt.LLSC(a, d, 1, "Iend_LE")
        assert m.tag == "Ist_LLSC"
        assert m.endness == "Iend_LE"
        assert m.result == 1
        assert m.addr.tmp == a.tmp
        assert m.storedata.tmp == d.tmp

    def test_irstmt_mbe(self):
        m = pyvex.IRStmt.MBE("Imbe_CancelReservation")
        assert m.event == "Imbe_CancelReservation"
        m.event = "Imbe_Fence"
        assert m.event == "Imbe_Fence"

    def test_irstmt_dirty(self):
        args = [pyvex.IRExpr.RdTmp.get_instance(i) for i in range(10)]
        m = pyvex.IRStmt.Dirty("test_dirty", pyvex.IRConst.U8(1), args, 15, "Ifx_None", 0, 1, 0)
        assert m.cee == "test_dirty"
        assert isinstance(m.guard, pyvex.IRConst.U8)
        assert m.tmp == 15
        assert m.mFx == "Ifx_None"
        assert m.nFxState == 0

        for n, a in enumerate(m.args):
            assert a.tmp == args[n].tmp

    def test_irstmt_exit(self):
        self.assertRaises(Exception, pyvex.IRStmt.Exit)

        g = pyvex.IRExpr.RdTmp.get_instance(123)
        d = pyvex.IRConst.U32(456)

        m = pyvex.IRStmt.Exit(g, d, "Ijk_Ret", 10)
        assert m.tag == "Ist_Exit"
        assert m.jumpkind == "Ijk_Ret"
        assert m.offsIP == 10
        assert m.guard.tmp == g.tmp
        assert m.dst.value == d.value

    ##################
    ### IRRegArray ###
    ##################

    def test_irregarray(self):
        m = pyvex.IRRegArray(10, "Ity_I64", 20)

        assert m.nElems == 20
        assert m.elemTy == "Ity_I64"
        assert m.base == 10

    ################
    ### IRConst.s ###
    ################

    def helper_const_subtype(self, subtype, tag, value):
        print("Testing %s" % tag)
        self.assertRaises(Exception, subtype)

        c = subtype(value)
        assert c.tag == tag
        assert c.value == value

        d = subtype(value - 1)
        e = subtype(value)
        assert c.value == e.value
        assert e.value == c.value
        self.assertNotEqual(c.value, d.value)
        self.assertNotEqual(d.value, c.value)
        self.assertNotEqual(c.value, "test")

        # TODO: actually check value
        assert c.type == d.type

    def test_irconst(self):
        self.helper_const_subtype(pyvex.IRConst.U1, "Ico_U1", 1)
        self.helper_const_subtype(pyvex.IRConst.U8, "Ico_U8", 233)
        self.helper_const_subtype(pyvex.IRConst.U16, "Ico_U16", 39852)
        self.helper_const_subtype(pyvex.IRConst.U32, "Ico_U32", 3442312356)
        self.helper_const_subtype(pyvex.IRConst.U64, "Ico_U64", 823452334523623455)
        self.helper_const_subtype(pyvex.IRConst.F32, "Ico_F32", 13453.234375)
        self.helper_const_subtype(pyvex.IRConst.F32i, "Ico_F32i", 3442312356)
        self.helper_const_subtype(pyvex.IRConst.F64, "Ico_F64", 13453.234525)
        self.helper_const_subtype(pyvex.IRConst.F64i, "Ico_F64i", 823457234523623455)
        self.helper_const_subtype(pyvex.IRConst.V128, "Ico_V128", 39852)
        self.helper_const_subtype(pyvex.IRConst.V256, "Ico_V256", 3442312356)

    ###################
    ### Expressions ###
    ###################

    def test_irexpr_binder(self):
        # binder doesn't work statically, but hopefully we should
        # never see it, anyways
        return
        # m = pyvex.IRExpr.Binder(1534252)
        # assert m.binder == 1534252

    def test_irexpr_geti(self):
        r = pyvex.IRRegArray(10, "Ity_I64", 20)
        i = pyvex.IRExpr.RdTmp.get_instance(5)
        m = pyvex.IRExpr.GetI(r, i, 2)
        assert m.description.base == 10
        assert m.index.tmp == 5
        assert m.bias == 2

        self.assertRaises(Exception, pyvex.IRExpr.GetI)

    def test_irexpr_rdtmp(self):
        m = pyvex.IRExpr.RdTmp.get_instance(123)
        assert m.tag == "Iex_RdTmp"
        assert m.tmp == 123

        irsb = pyvex.IRSB(b"\x90\x5d\xc3", mem_addr=0x0, arch=pyvex.ARCH_AMD64)
        print("TMP:", irsb.next.tmp)

    def test_irexpr_get(self):
        m = pyvex.IRExpr.Get(0, "Ity_I64")
        assert m.type == "Ity_I64"

        self.assertRaises(Exception, pyvex.IRExpr.Get)

    def test_irexpr_qop(self):
        a = pyvex.IRExpr.Get(0, "Ity_I64")
        b = pyvex.IRExpr.Get(184, "Ity_I64")
        c = pyvex.IRExpr.RdTmp.get_instance(1)
        d = pyvex.IRExpr.RdTmp.get_instance(2)
        op = "Iop_QAdd32S"

        m = pyvex.IRExpr.Qop(op, [a, b, c, d])

        assert m.op == op
        assert m.args[1].type == b.type

        assert len(m.args) == 4
        assert m.args[2].tmp == c.tmp

    def test_irexpr_triop(self):
        a = pyvex.IRExpr.Get(0, "Ity_I64")
        b = pyvex.IRExpr.Get(184, "Ity_I64")
        c = pyvex.IRExpr.RdTmp.get_instance(1)
        op = "Iop_MAddF64"

        m = pyvex.IRExpr.Triop(op, [a, b, c])

        assert m.op == op
        assert m.args[1].type == b.type

        assert len(m.args) == 3
        assert m.args[2].tmp == c.tmp

    def test_irexpr_binop(self):
        a = pyvex.IRExpr.Get(0, "Ity_I64")
        c = pyvex.IRExpr.RdTmp.get_instance(1)
        op = "Iop_Add64"

        m = pyvex.IRExpr.Binop(op, [a, c])

        assert m.op == op
        assert m.args[1].tmp == c.tmp

        assert len(m.args) == 2
        assert m.args[1].tmp == c.tmp

    def test_irexpr_unop(self):
        a = pyvex.IRExpr.Get(0, "Ity_I64")
        op = "Iop_Add64"

        m = pyvex.IRExpr.Unop(op, [a])

        assert m.op == op
        assert len(m.args) == 1
        assert m.args[0].offset == a.offset

    def test_irexpr_load(self):
        a = pyvex.IRExpr.Get(0, "Ity_I64")
        e = "Iend_LE"
        t = "Ity_I64"

        m = pyvex.IRExpr.Load(e, t, a)

        assert m.endness == e
        assert m.type == t

    def test_irexpr_const(self):
        u1 = pyvex.IRConst.U1(1)
        f64 = pyvex.IRConst.F64(1.123)

        ue = pyvex.IRExpr.Const(u1)
        _ = pyvex.IRExpr.Const(f64)

        assert ue.con.value == u1.value
        assert ue.con.value != f64.value

    def test_irexpr_ite(self):
        a = pyvex.IRExpr.Get(0, "Ity_I64")
        iffalse = pyvex.IRExpr.RdTmp.get_instance(1)
        iftrue = pyvex.IRExpr.Const(pyvex.IRConst.U8(200))

        m = pyvex.IRExpr.ITE(a, iffalse, iftrue)

        assert m.iftrue.con.value == iftrue.con.value

    def test_irexpr_ccall(self):
        callee = pyvex.IRCallee(3, "test_name", 0xFFFFFF)
        args = [pyvex.IRExpr.RdTmp.get_instance(i) for i in range(10)]

        m = pyvex.IRExpr.CCall("Ity_I64", callee, args)

        assert len(m.args) == len(args)
        assert m.ret_type == "Ity_I64"

        for n, a in enumerate(m.args):
            assert a.tmp == args[n].tmp

        m = pyvex.IRExpr.CCall(callee, "Ity_I64", ())
        assert len(m.args) == 0


if __name__ == "__main__":
    unittest.main()



import pyvex


##########################
### ARM Postprocessing ###
##########################
def test_arm_postprocess_call():
    for i in range(3):
        # Thumb

        # push  {r7}
        # add   r7, sp, #0
        # mov.w r1, #6
        # mov   r0, pc
        # add.w lr, r0, r1
        # b.w   10408
        irsb = pyvex.IRSB(
            data=(b"\x80\xb4" b"\x00\xaf" b"\x4f\xf0\x06\x01" b"\x78\x46" b"\x00\xeb\x01\x0e" b"\xff\xf7\xec\xbf"),
            mem_addr=0x1041F,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=6,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # mov   lr, pc
        # b.w   10408
        irsb = pyvex.IRSB(
            data=(b"\xfe\x46" b"\xe9\xe7"),
            mem_addr=0x10431,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=2,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # add   r2, pc, #0
        # add.w lr, r2, #4
        # ldr.w pc, [pc, #52]
        irsb = pyvex.IRSB(
            data=(b"\x00\xa2" b"\x02\xf1\x06\x0e" b"\xdf\xf8\x34\xf0"),
            mem_addr=0x10435,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=3,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # ldr   r0, [pc, #48]
        # mov   r1, pc
        # add.w r2, r1, #4
        # add.w r3, r2, #4
        # add.w r4, r3, #4
        # add.w lr, r4, #4
        # mov   pc, r0
        irsb = pyvex.IRSB(
            data=(
                b"\x0c\x48"
                b"\x79\x46"
                b"\x01\xf1\x04\x02"
                b"\x02\xf1\x04\x03"
                b"\x03\xf1\x04\x04"
                b"\x04\xf1\x04\x0e"
                b"\x87\x46"
            ),
            mem_addr=0x1043F,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=7,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # eor.w r0, r0, r0
        # mov   lr, pc
        # b.n   10460
        irsb = pyvex.IRSB(
            data=(b"\x80\xea\x00\x00" b"\x86\x46" b"\x01\xe0"),
            mem_addr=0x10455,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=3,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Boring"

        # Thumb compiled with optimizations (gcc -O2)

        # mov.w r1, #6
        # mov   r0, pc
        # add.w lr, r0, r1
        # b.w   104bc
        irsb = pyvex.IRSB(
            data=(b"\x4f\xf0\x06\x01" b"\x78\x46" b"\x00\xeb\x01\x0e" b"\x00\xf0\xc5\xb8"),
            mem_addr=0x10325,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=4,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # ldr   r0, [pc, #56]
        # mov   r1, pc
        # add.w r2, r1, #4
        # add.w r3, r2, #4
        # add.w r4, r3, #4
        # add.w lr, r4, #4
        # mov   pc, r0
        irsb = pyvex.IRSB(
            data=(
                b"\x0e\x48"
                b"\x79\x46"
                b"\x01\xf1\x04\x02"
                b"\x02\xf1\x04\x03"
                b"\x03\xf1\x04\x04"
                b"\x04\xf1\x04\x0e"
                b"\x87\x46"
            ),
            mem_addr=0x10333,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=7,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # add   r2, pc, #0
        # add.w lr, r2, #6
        # ldr.w pc, [pc, #28]
        irsb = pyvex.IRSB(
            data=(b"\x00\xa2" b"\x02\xf1\x06\x0e" b"\xdf\xf8\x1c\xf0"),
            mem_addr=0x10349,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=3,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # mov   lr, pc
        # b.w   104bc
        irsb = pyvex.IRSB(
            data=(b"\xfe\x46" b"\xb2\xe0"),
            mem_addr=0x10353,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=2,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # eor.w r0, r0, r0
        # mov   lr, pc
        # b.n   10362
        irsb = pyvex.IRSB(
            data=(b"\x80\xea\x00\x00" b"\x86\x46" b"\x01\xe0"),
            mem_addr=0x10357,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=3,
            bytes_offset=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Boring"

        # ARM compiled with optimizations (gcc -O2)

        # mov   r1, #4
        # mov   r0, pc
        # add   lr, r0, r1
        # ldr   pc, [pc, #56]
        irsb = pyvex.IRSB(
            data=(b"\x04\x10\xa0\xe3" b"\x0f\x00\xa0\xe1" b"\x01\xe0\x80\xe0" b"\x38\xf0\x9f\xe5"),
            mem_addr=0x10298,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=4,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # add   r1, pc, #0
        # add   r2, r1, #4
        # add   r3, r2, #4
        # add   r4, r3, #4
        # add   lr, r4, #4
        # b     10414
        irsb = pyvex.IRSB(
            data=(
                b"\x00\x10\x8f\xe2"
                b"\x04\x20\x81\xe2"
                b"\x04\x30\x82\xe2"
                b"\x04\x40\x83\xe2"
                b"\x04\xe0\x84\xe2"
                b"\x54\x00\x00\xea"
            ),
            mem_addr=0x102A8,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=6,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # mov   lr, pc
        # b     10414
        irsb = pyvex.IRSB(
            data=(b"\x0f\xe0\xa0\xe1" b"\x52\x00\x00\xea"),
            mem_addr=0x102C0,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=2,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # eor   r0, r0, r0
        # mov   lr, r0
        # b     102d8
        irsb = pyvex.IRSB(
            data=(b"\x00\x00\x20\xe0" b"\x00\xe0\xa0\xe1" b"\x00\x00\x00\xea"),
            mem_addr=0x102C8,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=3,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Boring"

        # ARM

        # push  {fp}
        # add   fp, sp, #0
        # mov   r1, #4
        # mov   r0, pc
        # add   lr, r0, r1
        # ldr   pc, [pc, #68]
        irsb = pyvex.IRSB(
            data=(
                b"\x04\xb0\x2d\xe5"
                b"\x00\xb0\x8d\xe2"
                b"\x04\x10\xa0\xe3"
                b"\x0f\x00\xa0\xe1"
                b"\x01\xe0\x80\xe0"
                b"\x44\xf0\x9f\xe5"
            ),
            mem_addr=0x103E8,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=6,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # add   r1, pc, #0
        # add   r2, r1, #4
        # add   r3, r2, #4
        # add   r4, r3, #4
        # add   lr, r4, #4
        # b     103c4
        irsb = pyvex.IRSB(
            data=(
                b"\x00\x10\x8f\xe2"
                b"\x04\x20\x81\xe2"
                b"\x04\x30\x82\xe2"
                b"\x04\x40\x83\xe2"
                b"\x04\xe0\x84\xe2"
                b"\x54\xff\xff\xea"
            ),
            mem_addr=0x10400,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=6,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # mov   lr, pc
        # b     103c4
        irsb = pyvex.IRSB(
            data=(b"\x0f\xe0\xa0\xe1" b"\xe8\xff\xff\xea"),
            mem_addr=0x10418,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=2,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"

        # eor   r0, r0, r0
        # mov   lr, r0
        # b     10430
        irsb = pyvex.IRSB(
            data=(b"\x00\x00\x20\xe0" b"\x00\xe0\xa0\xe1" b"\x00\x00\x00\xea"),
            mem_addr=0x10420,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=3,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Boring"

        # From a "real thing" compiled with armc
        # ARM:
        #
        irsb = pyvex.IRSB(
            data=(
                b"H\x10\x9b\xe5"
                b"\x0b\x00\xa0\xe1"
                b"\x04 \x91\xe5"
                b"\x04\xe0\x8f\xe2"
                b"\x01\x10\x82\xe0"
                b"\x01\xf0\xa0\xe1"
            ),
            mem_addr=0x264B4C,
            arch=pyvex.ARCH_ARM_LE,
            num_inst=6,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Call"


def test_arm_postprocess_ret():
    for i in range(3):
        # e91ba8f0
        # ldmdb  R11, {R4,R11,SP,PC}
        irsb = pyvex.IRSB(
            data=b"\xe9\x1b\xa8\xf0",
            mem_addr=0xED4028,
            arch=pyvex.ARCH_ARM_BE_LE,
            num_inst=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Ret"

        # e91badf0
        # ldmdb  R11, {R4-R8,R10,R11,SP,PC}
        irsb = pyvex.IRSB(
            data=b"\xe9\x1b\xa8\xf0",
            mem_addr=0x4D4028,
            arch=pyvex.ARCH_ARM_BE_LE,
            num_inst=1,
            opt_level=i,
        )
        assert irsb.jumpkind == "Ijk_Ret"

        # 00a89de8
        # ldmfd SP, {R11,SP,PC}
        # Fixed by Fish in the VEX fork, commit 43c78f608490f9a5c71c7fca87c04759c1b93741
        irsb = pyvex.IRSB(
            data=b"\x00\xa8\x9d\xe8",
            mem_addr=0xC800B57C,
            arch=pyvex.ARCH_ARM_BE,
            num_inst=1,
            opt_level=1,
        )
        assert irsb.jumpkind == "Ijk_Ret"


if __name__ == "__main__":
    test_arm_postprocess_call()
    test_arm_postprocess_ret()



class ConstVal:
    """
    A constant value object. Indicates a constant value assignment to a VEX tmp variable.

    :ivar tmp:          The tmp variable being assigned to.
    :ivar value:        The value of the tmp variable.
    :ivar stmt_idx:     The IRSB statement index containing the data access
    """

    __slots__ = (
        "tmp",
        "value",
        "stmt_idx",
    )

    def __init__(self, tmp: int, value: int, stmt_idx: int):
        self.tmp = tmp
        self.value = value
        self.stmt_idx = stmt_idx

    def __repr__(self):
        return f"<ConstVal {self.tmp} = {self.value:#x} @ {self.stmt_idx}>"

    @classmethod
    def from_c(cls, r):
        return cls(r.tmp, r.value, r.stmt_idx)



# pylint: disable=missing-class-docstring
import unittest

import pyvex


class Tests(unittest.TestCase):
    def test_x86_aam(self):
        irsb = pyvex.lift(b"\xd4\x0b", 0, pyvex.ARCH_X86)
        self.assertEqual(irsb.jumpkind, "Ijk_Boring")
        self.assertEqual(irsb.size, 2)

    def test_x86_aad(self):
        irsb = pyvex.lift(b"\xd5\x0b", 0, pyvex.ARCH_X86)
        self.assertEqual(irsb.jumpkind, "Ijk_Boring")
        self.assertEqual(irsb.size, 2)

    def test_x86_xgetbv(self):
        irsb = pyvex.lift(b"\x0f\x01\xd0", 0, pyvex.ARCH_X86)
        self.assertEqual(irsb.jumpkind, "Ijk_Boring")
        self.assertEqual(irsb.size, 3)

    def test_x86_rdmsr(self):
        irsb = pyvex.lift(b"\x0f\x32", 0, pyvex.ARCH_X86)
        self.assertEqual(irsb.jumpkind, "Ijk_Boring")
        self.assertEqual(irsb.size, 2)


if __name__ == "__main__":
    unittest.main()



import logging
import threading
from typing import TYPE_CHECKING

from pyvex.errors import LiftingException
from pyvex.native import ffi, pvc
from pyvex.types import CLiftSource, LibvexArch

from .lift_function import Lifter

log = logging.getLogger("pyvex.lifting.libvex")

_libvex_lock = threading.Lock()

LIBVEX_SUPPORTED_ARCHES = {
    "X86",
    "AMD64",
    "MIPS32",
    "MIPS64",
    "ARM",
    "ARMEL",
    "ARMHF",
    "ARMCortexM",
    "AARCH64",
    "PPC32",
    "PPC64",
    "S390X",
    "RISCV64",
}

VEX_MAX_INSTRUCTIONS = 99
VEX_MAX_BYTES = 5000


class VexRegisterUpdates:
    VexRegUpd_INVALID = 0x700
    VexRegUpdSpAtMemAccess = 0x701
    VexRegUpdUnwindregsAtMemAccess = 0x702
    VexRegUpdAllregsAtMemAccess = 0x703
    VexRegUpdAllregsAtEachInsn = 0x704
    VexRegUpdLdAllregsAtEachInsn = 0x705


class LibVEXLifter(Lifter):
    __slots__ = ()

    REQUIRE_DATA_C = True

    @staticmethod
    def get_vex_log():
        return bytes(ffi.buffer(pvc.msg_buffer, pvc.msg_current_size)).decode() if pvc.msg_buffer != ffi.NULL else None

    def _lift(self):
        if TYPE_CHECKING:
            assert isinstance(self.irsb.arch, LibvexArch)
            assert isinstance(self.data, CLiftSource)
        try:
            _libvex_lock.acquire()

            pvc.log_level = log.getEffectiveLevel()
            vex_arch = getattr(pvc, self.irsb.arch.vex_arch, None)
            assert vex_arch is not None

            if self.bytes_offset is None:
                self.bytes_offset = 0

            if self.max_bytes is None or self.max_bytes > VEX_MAX_BYTES:
                max_bytes = VEX_MAX_BYTES
            else:
                max_bytes = self.max_bytes

            if self.max_inst is None or self.max_inst > VEX_MAX_INSTRUCTIONS:
                max_inst = VEX_MAX_INSTRUCTIONS
            else:
                max_inst = self.max_inst

            strict_block_end = self.strict_block_end
            if strict_block_end is None:
                strict_block_end = True

            if self.cross_insn_opt:
                px_control = VexRegisterUpdates.VexRegUpdUnwindregsAtMemAccess
            else:
                px_control = VexRegisterUpdates.VexRegUpdLdAllregsAtEachInsn

            self.irsb.arch.vex_archinfo["hwcache_info"]["caches"] = ffi.NULL
            lift_r = pvc.vex_lift(
                vex_arch,
                self.irsb.arch.vex_archinfo,
                self.data + self.bytes_offset,
                self.irsb.addr,
                max_inst,
                max_bytes,
                self.opt_level,
                self.traceflags,
                self.allow_arch_optimizations,
                strict_block_end,
                1 if self.collect_data_refs else 0,
                1 if self.load_from_ro_regions else 0,
                1 if self.const_prop else 0,
                px_control,
                self.bytes_offset,
            )
            log_str = self.get_vex_log()
            if lift_r == ffi.NULL:
                raise LiftingException("libvex: unknown error" if log_str is None else log_str)
            else:
                if log_str is not None:
                    log.debug(log_str)

            self.irsb._from_c(lift_r, skip_stmts=self.skip_stmts)
            if self.irsb.size == 0:
                log.debug("raising lifting exception")
                raise LiftingException("libvex: could not decode any instructions @ 0x%x" % self.addr)
        finally:
            _libvex_lock.release()
            self.irsb.arch.vex_archinfo["hwcache_info"]["caches"] = None



import hashlib
import os
import pickle
import sys
import tempfile
from typing import Any

import cffi

from .vex_ffi import ffi_str as _ffi_str

ffi = cffi.FFI()


def _locate_lib(module: str, library: str) -> str:
    """
    Attempt to find a native library without using pkg_resources, and only fall back to pkg_resources upon failures.
    This is because "import pkg_resources" is slow.

    :return:    The full path of the native library.
    """
    base_dir = os.path.dirname(__file__)
    attempt = os.path.join(base_dir, library)
    if os.path.isfile(attempt):
        return attempt

    import pkg_resources  # pylint:disable=import-outside-toplevel

    return pkg_resources.resource_filename(module, os.path.join("lib", library))


def _parse_ffi_str():
    hash_ = hashlib.md5(_ffi_str.encode("utf-8")).hexdigest()
    cache_location = os.path.join(tempfile.gettempdir(), f"pyvex_ffi_parser_cache.{hash_}")

    if os.path.isfile(cache_location):
        # load the cache
        with open(cache_location, "rb") as f:
            cache = pickle.loads(f.read())
        ffi._parser._declarations = cache["_declarations"]
        ffi._parser._int_constants = cache["_int_constants"]
    else:
        ffi.cdef(_ffi_str)
        # cache the result
        cache = {
            "_declarations": ffi._parser._declarations,
            "_int_constants": ffi._parser._int_constants,
        }
        with open(cache_location, "wb") as f:
            f.write(pickle.dumps(cache))


def _find_c_lib():
    # Load the c library for calling into VEX
    if sys.platform in ("win32", "cygwin"):
        library_file = "pyvex.dll"
    elif sys.platform == "darwin":
        library_file = "libpyvex.dylib"
    else:
        library_file = "libpyvex.so"

    pyvex_path = _locate_lib(__name__, os.path.join("lib", library_file))
    # parse _ffi_str and use cache if possible
    _parse_ffi_str()
    # RTLD_GLOBAL used for sim_unicorn.so
    lib = ffi.dlopen(pyvex_path)
    if not lib.vex_init():
        raise ImportError("libvex failed to initialize")
    # this looks up all the definitions (wtf)
    dir(lib)
    return lib


pvc: Any = _find_c_lib()  # This should be properly typed, but this seems non trivial



import pyvex


def test_mips32_unconditional_jumps():
    # 0040000c: 10000002 ; <input:28> beq $zero, $zero, LABEL_ELSE_IF
    # 00400010: 00000000 ; <input:31> sll $zero, $zero, 0
    # 00400014: 08100012 ; <input:34> j LABEL_DONE
    # 00400018: <LABEL_ELSE_IF> ; <input:37> LABEL_ELSE_IF:
    irsb = pyvex.IRSB(
        data=(b"\x10\x00\x00\x02" b"\x00\x00\x00\x00"),
        mem_addr=0x40000C,
        arch=pyvex.ARCH_MIPS32_BE,
        num_inst=2,
        opt_level=0,
    )
    assert type(irsb.next) is pyvex.expr.Const
    assert irsb.next.con.value == 0x400018


if __name__ == "__main__":
    test_mips32_unconditional_jumps()



import logging

import bitstring

from pyvex.lifting.util import JumpKind, Type
from pyvex.lifting.util.instr_helper import Instruction, ParseError
from pyvex.lifting.util.lifter_helper import GymratLifter
from pyvex.types import Arch

log = logging.getLogger(__name__)


class ARMInstruction(Instruction):  # pylint: disable=abstract-method
    # NOTE: WARNING: There is no CPSR in VEX's ARM implementation
    # You must use straight nasty hacks instead.

    # NOTE 2: Something is goofy w/r/t archinfo and VEX; cc_op3 is used in ccalls, but there's
    # no cc_op3 in archinfo, angr itself uses cc_depn instead.  We do the same.

    def match_instruction(self, data, bitstrm):
        """
        ARM Instructions are pretty dense, so let's do what we can to weed them out
        """
        if "c" not in data or data["c"] == "1111":
            raise ParseError("Invalid ARM Instruction")

    def get_N(self):
        cc_op = self.get("cc_op", Type.int_32)
        cc_dep1 = self.get("cc_dep1", Type.int_32)
        cc_dep2 = self.get("cc_dep2", Type.int_32)
        cc_depn = self.get("cc_ndep", Type.int_32)
        return self.ccall(Type.int_32, "armg_calculate_flag_n", [cc_op, cc_dep1, cc_dep2, cc_depn])

    def get_C(self):
        cc_op = self.get("cc_op", Type.int_32)
        cc_dep1 = self.get("cc_dep1", Type.int_32)
        cc_dep2 = self.get("cc_dep2", Type.int_32)
        cc_depn = self.get("cc_ndep", Type.int_32)
        return self.ccall(Type.int_32, "armg_calculate_flag_c", [cc_op, cc_dep1, cc_dep2, cc_depn])

    def get_V(self):
        cc_op = self.get("cc_op", Type.int_32)
        cc_dep1 = self.get("cc_dep1", Type.int_32)
        cc_dep2 = self.get("cc_dep2", Type.int_32)
        cc_depn = self.get("cc_ndep", Type.int_32)
        return self.ccall(Type.int_32, "armg_calculate_flag_v", [cc_op, cc_dep1, cc_dep2, cc_depn])

    def get_Z(self):
        cc_op = self.get("cc_op", Type.int_32)
        cc_dep1 = self.get("cc_dep1", Type.int_32)
        cc_dep2 = self.get("cc_dep2", Type.int_32)
        cc_depn = self.get("cc_ndep", Type.int_32)
        return self.ccall(Type.int_32, "armg_calculate_flag_z", [cc_op.rdt, cc_dep1.rdt, cc_dep2.rdt, cc_depn.rdt])

    def evaluate_condition(self):
        # condition codes should be in 'c'
        cond = self.data["c"]
        if cond == "0000":
            # equal, z set
            return self.get_Z() == 1
        elif cond == "0001":
            # not equal, Z clear
            return self.get_Z() == 0
        elif cond == "0010":
            # Carry, C set
            return self.get_C() == 1
        elif cond == "0011":
            # Carry Clear, C clear
            return self.get_C() == 0
        elif cond == "0100":
            # MI / neagative / N set
            return self.get_N() == 1
        elif cond == "0101":
            # PL / plus / positive / N clear
            return self.get_N() == 0
        elif cond == "0110":
            # VS / V set / Overflow
            return self.get_V() == 1
        elif cond == "0111":
            # VC / V Clear / no overflow
            return self.get_V() == 0
        elif cond == "1000":
            # Hi / unsigned higher / C set, Z clear
            return (self.get_C() == 1) & (self.get_Z() == 0)
        elif cond == "1001":
            # LS / C clear, Z set
            return (self.get_C() == 0) & (self.get_Z() == 1)
        elif cond == "1011":
            # LT / Less than / N != V
            return self.get_N() != self.get_V()
        elif cond == "1100":
            # GT / greater than / Z clear and (n == v)
            return (self.get_Z() == 1) & (self.get_N() != self.get_V())
        elif cond == "1101":
            # LE / less than or equal to / Z set OR (N != V)
            return (self.get_Z() == 1) | (self.get_N() != self.get_V())
        else:
            # No condition
            return None

    def _load_le_instr(self, bitstream: bitstring.ConstBitStream, numbits: int) -> str:
        # THUMB mode instructions swap endianness every two bytes!
        if (self.addr & 1) == 1 and numbits > 16:
            chunk = ""
            oldpos = bitstream.pos
            try:
                for _ in range(0, numbits, 16):
                    chunk += bitstring.Bits(uint=bitstream.peek("uintle:%d" % 16), length=16).bin
                    bitstream.pos += 16
            finally:
                bitstream.pos = oldpos
            return chunk
        return super()._load_le_instr(bitstream, numbits)


class Instruction_MRC(ARMInstruction):
    name = "MRC"
    bin_format = "cccc1110CCC1nnnnddddppppOOOOOOOO"
    # 11101110000100010001111100010000
    # c = cond
    # C = Coprocessor operation mode
    # d = CPd
    # O = Offset
    # p = CP#

    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO at least look at the conditionals
        # TODO Clobber the dst reg of MCR
        # TODO maybe treat coproc regs as simple storage (even though they are very much not)
        log.debug("Ignoring MRC instruction at %#x.", self.addr)


class Instruction_MCR(ARMInstruction):
    name = "MCR"
    bin_format = "cccc1110CCC0nnnnddddppppOOOOOOOO"
    # 11101110000000010000111100010000
    # c = cond
    # C = Coprocessor operation mode
    # d = CPd
    # O = Offset
    # p = CP#

    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO at least look at the conditionals
        # TODO Clobber the dst reg of MCR
        # TODO maybe treat coproc regs as simple storage (even though they are very much not)
        log.debug("Ignoring MCR instruction at %#x.", self.addr)


class Instruction_MSR(ARMInstruction):
    name = "MSR"
    bin_format = "cccc00i10d10xxxj1111ssssssssssss"
    #             11100011001000011111000010010001
    #             11100001011011111111000000000001

    def compute_result(self):  # pylint: disable=arguments-differ
        log.debug(
            "Ignoring MSR instruction at %#x. VEX cannot support this instruction. "
            "See pyvex/lifting/gym/arm_spotter.py",
            self.addr,
        )


class Instruction_MRS(ARMInstruction):
    name = "MRS"
    bin_format = "cccc00010s001111dddd000000000000"

    def compute_result(self):  # pylint: disable=arguments-differ
        log.debug(
            "Ignoring MRS instruction at %#x. VEX cannot support this instruction. "
            "See pyvex/lifting/gym/arm_spotter.py",
            self.addr,
        )


class Instruction_STM(ARMInstruction):
    name = "STM"
    bin_format = "cccc100pu1w0bbbbrrrrrrrrrrrrrrrr"

    def match_instruction(self, data, bitstrm):
        # If we don't push anything, that's not real
        if int(data["r"]) == 0:
            raise ParseError("Invalid STM instruction")
        return True

    def compute_result(self):  # pylint: disable=arguments-differ
        log.debug(
            "Ignoring STMxx ^ instruction at %#x. This mode is not implemented by VEX! "
            "See pyvex/lifting/gym/arm_spotter.py",
            self.addr,
        )


class Instruction_LDM(ARMInstruction):
    name = "LDM"
    bin_format = "cccc100PU1W1bbbbrrrrrrrrrrrrrrrr"

    def match_instruction(self, data, bitstrm):
        # If we don't push anything, that's not real
        if int(data["r"]) == 0:
            raise ParseError("Invalid LDM instruction")
        return True

    def compute_result(self):  # pylint: disable=arguments-differ
        # test if PC will be set. If so, the jumpkind of this block should be Ijk_Ret
        log.debug("Spotting an LDM instruction at %#x.  This is not fully tested.  Prepare for errors.", self.addr)

        src_n = f"r{int(self.data['b'], 2)}"
        src = self.get(src_n, Type.int_32)

        for reg_num, bit in enumerate(self.data["r"]):
            reg_num = 15 - reg_num
            if bit == "1":
                if self.data["P"] == "1":
                    if self.data["U"] == "0":
                        src += 4
                    else:
                        src -= 4
                val = self.load(src, Type.int_32)
                self.put(val, f"r{reg_num}")
                if self.data["P"] == "0":
                    if self.data["U"] == "0":
                        src += 4
                    else:
                        src -= 4
                # If we touch PC, we're doing a RET!
                if reg_num == 15 and bit == "1":
                    cond = self.evaluate_condition()
                    if cond is not None:
                        self.jump(cond, val, JumpKind.Ret)
                    else:
                        self.jump(None, val, JumpKind.Ret)
        # Write-back
        if self.data["W"] == "1":
            self.put(src, src_n)


class Instruction_STC(ARMInstruction):
    name = "STC"
    bin_format = "cccc110PUNW0nnnnddddppppOOOOOOOO"

    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        log.debug("Ignoring STC instruction at %#x.", self.addr)


class Instruction_STC_THUMB(ARMInstruction):
    name = "STC"
    bin_format = "111c110PUNW0nnnnddddppppOOOOOOOO"

    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        log.debug("Ignoring STC instruction at %#x.", self.addr)


class Instruction_LDC(ARMInstruction):
    name = "LDC"
    bin_format = "cccc110PUNW1nnnnddddppppOOOOOOOO"

    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        # TODO Clobber the dest reg of LDC
        # TODO Maybe clobber the dst reg of CDP, if we're really adventurous
        log.debug("Ignoring LDC instruction at %#x.", self.addr)


class Instruction_LDC_THUMB(ARMInstruction):
    name = "LDC"
    bin_format = "111c110PUNW1nnnnddddppppOOOOOOOO"

    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        # TODO Clobber the dest reg of LDC
        # TODO Maybe clobber the dst reg of CDP, if we're really adventurous
        log.debug("Ignoring LDC instruction at %#x.", self.addr)


class Instruction_CDP(Instruction):
    name = "CDP"
    bin_format = "cccc1110oooonnnnddddppppPPP0mmmm"
    # c = cond
    # d = CPd
    # O = Offset
    # p = CP#

    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO At least look at the conditionals
        # TODO Maybe clobber the dst reg of CDP, if we're really adventurous
        log.debug("Ignoring CDP instruction at %#x.", self.addr)


##
## Thumb! (ugh)
##


class ThumbInstruction(Instruction):  # pylint: disable=abstract-method
    def mark_instruction_start(self):
        self.irsb_c.imark(self.addr - 1, self.bytewidth, 1)


class Instruction_tCPSID(ThumbInstruction):
    name = "CPSID"
    bin_format = "101101x0011x0010"

    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO haha lol yeah right
        log.debug("[thumb] Ignoring CPS instruction at %#x.", self.addr)


class Instruction_tMSR(ThumbInstruction):
    name = "tMSR"
    bin_format = "10x0mmmmxxxxxxxx11110011100Rrrrr"

    def compute_result(self):  # pylint: disable=arguments-differ
        dest_spec_reg = int(self.data["x"], 2)
        src_reg = f"r{int(self.data['r'], 2)}"

        # If 0, do not write the SPSR
        if self.data["R"] == "0":
            if dest_spec_reg == 8:  # msp
                src = self.get(src_reg, Type.int_32)
                self.put(src, "sp")
            elif dest_spec_reg == 16:  # primask
                src = self.get(src_reg, Type.int_32)
                self.put(src, "primask")
            else:
                log.debug(
                    "[thumb] FIXME: tMSR at %#x is writing into an unsupported special register %#x. "
                    "Ignoring the instruction.",
                    self.addr,
                    dest_spec_reg,
                )
        else:
            log.debug("[thumb] tMSR at %#x is writing SPSR. Ignoring the instruction. FixMe.", self.addr)
        log.debug(
            "[thumb] Spotting an tMSR instruction at %#x.  This is not fully tested.  Prepare for errors.", self.addr
        )


class Instruction_tMRS(ThumbInstruction):
    name = "tMRS"
    bin_format = "10x0mmmmxxxxxxxx11110011111Rrrrr"

    def compute_result(self):  # pylint: disable=arguments-differ
        spec_reg = int(self.data["x"], 2)
        dest_reg = f"r{int(self.data['m'], 2)}"

        # Reading from CPSR
        if self.data["R"] == "0":
            # See special registers constants here:
            # https://github.com/aquynh/capstone/blob/45bec1a691e455b864f7e4d394711a467e5493dc/arch/ARM/ARMInstPrinter.c#L1654
            if spec_reg == 8:
                # We move the SP and call it a day.
                src = self.get("sp", Type.int_32)
                self.put(src, dest_reg)
            elif spec_reg == 16:
                src = self.get("primask", Type.int_32)
                self.put(src, dest_reg)
            else:
                log.debug(
                    "[thumb] FIXME: tMRS at %#x is using the unsupported special register %#x. "
                    "Ignoring the instruction.",
                    self.addr,
                    spec_reg,
                )
        else:
            log.debug("[thumb] tMRS at %#x is reading from SPSR. Ignoring the instruction. FixMe.", self.addr)
            log.debug("[thumb] Ignoring tMRS instruction at %#x.", self.addr)
        log.debug(
            "[thumb] Spotting an tMRS instruction at %#x.  This is not fully tested.  Prepare for errors.", self.addr
        )


class Instruction_tDMB(ThumbInstruction):
    name = "DMB"
    bin_format = "100011110101xxxx1111001110111111"

    def compute_result(self):  # pylint: disable=arguments-differ
        # TODO haha lol yeah right
        log.debug("[thumb] Ignoring DMB instruction at %#x.", self.addr)


class Instruction_WFI(ThumbInstruction):
    name = "WFI"
    bin_format = "10111111001a0000"
    # 1011111100110000

    def compute_result(self):  # pylint: disable=arguments-differ
        log.debug("[thumb] Ignoring WFI instruction at %#x.", self.addr)


class ARMSpotter(GymratLifter):
    arm_instrs = [
        Instruction_MRC,
        Instruction_MCR,
        Instruction_MSR,
        Instruction_MRS,
        Instruction_STM,
        Instruction_LDM,
        Instruction_STC,
        Instruction_LDC,
        Instruction_CDP,
    ]
    thumb_instrs = [
        Instruction_tCPSID,
        Instruction_tMSR,
        Instruction_tMRS,
        Instruction_WFI,
        Instruction_tDMB,
        Instruction_STC_THUMB,
        Instruction_LDC_THUMB,
    ]

    def __init__(self, arch: Arch, addr: int):
        super().__init__(arch, addr)
        self.thumb: bool = False

    def _lift(self):
        if self.irsb.addr & 1:
            # Thumb!
            self.instrs = self.thumb_instrs
            self.thumb = True
        else:
            self.instrs = self.arm_instrs
            self.thumb = False
        super()._lift()



from .aarch64_spotter import AARCH64Spotter
from .arm_spotter import ARMSpotter
from .x86_spotter import AMD64Spotter, X86Spotter

__all__ = ("ARMSpotter", "AARCH64Spotter", "X86Spotter", "AMD64Spotter")



import functools
from typing import Union

from pyvex.const import get_type_size
from pyvex.expr import Const, IRExpr, RdTmp

from .vex_helper import IRSBCustomizer, Type


def checkparams(rhstype=None):
    def decorator(fn):
        @functools.wraps(fn)
        def inner_decorator(self, *args, **kwargs):
            irsb_cs = {a.irsb_c for a in list(args) + [self] if isinstance(a, VexValue)}  # pylint: disable=no-member
            assert len(irsb_cs) == 1, "All VexValues must belong to the same irsb_c"
            args = list(args)
            for idx, arg in enumerate(args):
                if isinstance(arg, int):
                    thetype = rhstype if rhstype else self.ty
                    args[idx] = VexValue.Constant(self.irsb_c, arg, thetype)
                elif not isinstance(arg, VexValue):
                    raise Exception("Cannot convert param %s" % str(arg))
            args = tuple(args)
            return fn(self, *args, **kwargs)

        return inner_decorator

    return decorator


def vvifyresults(f):
    @functools.wraps(f)
    def decor(self, *args, **kwargs):
        returned = f(self, *args, **kwargs)
        assert isinstance(returned, RdTmp) or isinstance(returned, Const)
        return VexValue(self.irsb_c, returned)

    return decor


class VexValue:
    def __init__(self, irsb_c: "IRSBCustomizer", rdt: "Union[RdTmp, Const]", signed=False):
        self.irsb_c = irsb_c
        self.ty = self.irsb_c.get_type(rdt)
        self.rdt = rdt
        self.width = get_type_size(self.ty)
        self._is_signed = signed

    @property
    def value(self):
        if isinstance(self.rdt, Const):
            return self.rdt.con.value
        else:
            raise ValueError("Non-constant VexValue has no value property")

    @property
    def signed(self):
        return VexValue(self.irsb_c, self.rdt, True)

    @vvifyresults
    def widen_unsigned(self, ty):
        return self.irsb_c.op_widen_int_unsigned(self.rdt, ty)

    @vvifyresults
    def cast_to(self, ty, signed=False, high=False):
        return self.irsb_c.cast_to(self.rdt, ty, signed=signed, high=high)

    @vvifyresults
    def widen_signed(self, ty):
        return self.irsb_c.op_widen_int_signed(self.rdt, ty)

    @vvifyresults
    def narrow_high(self, ty):
        return self.irsb_c.op_narrow_int(self.rdt, ty, high_half=True)

    @vvifyresults
    def narrow_low(self, ty):
        return self.irsb_c.op_narrow_int(self.rdt, ty, high_half=False)

    # TODO at some point extend this to Vex nonconstants
    def __getitem__(self, idx):
        def getb(i):
            return VexValue(self.irsb_c, self.irsb_c.get_bit(self.rdt, i))

        def makeconstant(x):
            return VexValue.Constant(self.irsb_c, x, Type.int_8).rdt

        if not isinstance(idx, slice):
            actualindex = slice(idx).indices(self.width)[1]
            return getb(makeconstant(actualindex))
        else:
            return [getb(makeconstant(i)) for i in range(*idx.indices(self.width))]

    def __setitem__(self, idx, bval):
        setted = self.set_bit(idx, bval)
        self.__init__(setted.irsb_c, setted.rdt)

    @checkparams(rhstype=Type.int_8)
    @vvifyresults
    def set_bit(self, idx, bval):
        return self.irsb_c.set_bit(self.rdt, idx.rdt, bval.rdt)

    @checkparams()
    @vvifyresults
    def set_bits(self, idxsandvals):
        return self.irsb_c.set_bits(self.rdt, [(i.cast_to(Type.int_8).rdt, b.rdt) for i, b in idxsandvals])

    @checkparams()
    @vvifyresults
    def ite(self, iftrue, iffalse):
        onebitcond = self.cast_to(Type.int_1)
        return self.irsb_c.ite(onebitcond.rdt, iftrue.rdt, iffalse.rdt)

    @checkparams()
    @vvifyresults
    def sar(self, right):
        """
        `v.sar(r)` should do arithmetic shift right of `v` by `r`

        :param right:VexValue value to shift by
        :return: VexValue - result of a shift
        """
        return self.irsb_c.op_sar(self.rdt, right.rdt)

    @checkparams()
    @vvifyresults
    def __add__(self, right):
        return self.irsb_c.op_add(self.rdt, right.rdt)

    @checkparams()
    def __radd__(self, left):
        return self + left

    @checkparams()
    @vvifyresults
    def __sub__(self, right):
        return self.irsb_c.op_sub(self.rdt, right.rdt)

    @checkparams()
    def __rsub__(self, left):
        return left - self

    @checkparams()
    @vvifyresults
    def __div__(self, right):
        if self._is_signed:
            return self.irsb_c.op_sdiv(self.rdt, right.rdt)
        else:
            return self.irsb_c.op_udiv(self.rdt, right.rdt)

    @checkparams()
    def __rdiv__(self, left):
        return left // self

    @checkparams()
    def __floordiv__(self, right):  # Note: nonprimitive
        return self.__div__(right)

    @checkparams()
    def __rfloordiv__(self, left):
        return left // self

    @checkparams()
    def __truediv__(self, right):  # Note: nonprimitive
        return self / right

    @checkparams()
    def __rtruediv__(self, left):
        return left.__truediv__(self)

    @checkparams()
    @vvifyresults
    def __and__(self, right):
        return self.irsb_c.op_and(self.rdt, right.rdt)

    @checkparams()
    def __rand__(self, left):
        return left & self

    @checkparams()
    @vvifyresults
    def __eq__(self, right):
        return self.irsb_c.op_cmp_eq(self.rdt, right.rdt)

    @checkparams()
    @vvifyresults
    def __ne__(self, other):
        return self.irsb_c.op_cmp_ne(self.rdt, other.rdt)

    @checkparams()
    @vvifyresults
    def __invert__(self):
        return self.irsb_c.op_not(self.rdt)

    @checkparams()
    @vvifyresults
    def __le__(self, right):
        if self._is_signed:
            return self.irsb_c.op_cmp_sle(self.rdt, right.rdt)
        else:
            return self.irsb_c.op_cmp_ule(self.rdt, right.rdt)

    @checkparams()
    @vvifyresults
    def __gt__(self, other):
        if self._is_signed:
            return self.irsb_c.op_cmp_sgt(self.rdt, other.rdt)
        else:
            return self.irsb_c.op_cmp_ugt(self.rdt, other.rdt)

    @checkparams()
    @vvifyresults
    def __ge__(self, right):
        if self._is_signed:
            return self.irsb_c.op_cmp_sge(self.rdt, right.rdt)
        else:
            return self.irsb_c.op_cmp_uge(self.rdt, right.rdt)

    @checkparams(rhstype=Type.int_8)
    @vvifyresults
    def __lshift__(self, right):  # TODO put better type inference in irsb_c so we can have rlshift
        """
        logical shift left
        """
        return self.irsb_c.op_shl(self.rdt, right.rdt)

    @checkparams()
    @vvifyresults
    def __lt__(self, right):
        if self._is_signed:
            return self.irsb_c.op_cmp_slt(self.rdt, right.rdt)
        else:
            return self.irsb_c.op_cmp_ult(self.rdt, right.rdt)

    @checkparams()
    @vvifyresults
    def __mod__(self, right):  # Note: nonprimitive
        return self.irsb_c.op_mod(self.rdt, right.rdt)

    @checkparams()
    def __rmod__(self, left):
        return left % self

    @checkparams()
    @vvifyresults
    def __mul__(self, right):
        if self._is_signed:
            return self.irsb_c.op_smul(self.rdt, right.rdt)
        else:
            return self.irsb_c.op_umul(self.rdt, right.rdt)

    @checkparams()
    def __rmul__(self, left):
        return left * self

    @checkparams()
    @vvifyresults
    def __neg__(self):  # Note: nonprimitive
        if not self._is_signed:
            raise Exception("Number is unsigned, cannot change sign!")
        else:
            return self.rdt * -1

    @checkparams()
    @vvifyresults
    def __or__(self, right):
        return self.irsb_c.op_or(self.rdt, right.rdt)

    def __ror__(self, left):
        return self | left

    @checkparams()
    @vvifyresults
    def __pos__(self):
        return self

    @checkparams(rhstype=Type.int_8)
    @vvifyresults
    def __rshift__(self, right):
        """
        logical shift right
        """
        return self.irsb_c.op_shr(self.rdt, right.rdt)

    @checkparams()
    def __rlshift__(self, left):
        return left << self

    @checkparams()
    def __rrshift__(self, left):
        return left >> self

    @checkparams()
    @vvifyresults
    def __xor__(self, right):
        return self.irsb_c.op_xor(self.rdt, right.rdt)

    def __rxor__(self, left):
        return self ^ left

    @classmethod
    def Constant(cls, irsb_c, val, ty):
        """
        Creates a constant as a VexValue
        :param irsb_c: The IRSBCustomizer to use
        :param val: The value, as an integer
        :param ty: The type of the resulting VexValue
        :return: a VexValue
        """
        assert not (isinstance(val, VexValue) or isinstance(val, IRExpr))
        rdt = irsb_c.mkconst(val, ty)
        return cls(irsb_c, rdt)



import os

import pyvex
import pyvex.lifting
from pyvex.lifting import register
from pyvex.lifting.util import GymratLifter, Instruction, Type

test_location = str(os.path.join(os.path.dirname(os.path.realpath(__file__)), "../../binaries/tests"))


class Instruction_IMAGINARY(Instruction):
    bin_format = bin(0x0F0B)[2:].zfill(16)
    name = "IMAGINARY"

    def compute_result(self):
        a = self.constant(10, Type.int_27)
        b = self.constant(20, Type.int_27)
        a + b


class ImaginarySpotter(GymratLifter):
    instrs = [Instruction_IMAGINARY]


register(ImaginarySpotter, "X86")

basic_goal = """
IRSB {
   t0:Ity_I27

   00 | ------ IMark(0x1, 2, 0) ------
   01 | t0 = Add27((0xa :: Ity_I27),(0x14 :: Ity_I27))
   NEXT: PUT(eip) = 0x00000003; Ijk_Boring
}
"""


def test_basic():
    b = pyvex.block.IRSB(b"\x0f\x0b", 1, pyvex.ARCH_X86)
    assert str(b).strip() == basic_goal.strip()


def test_embedded():
    b = pyvex.block.IRSB(b"\x50" * 3 + b"\x0f\x0b" + b"\x50" * 6, 1, pyvex.ARCH_X86)
    for i, stmt in enumerate(b.statements):
        if type(stmt) is pyvex.stmt.IMark and stmt.addr == 0x4 and stmt.len == 2 and stmt.delta == 0:
            imaginary_trans_stmt = b.statements[i + 1]
            assert type(imaginary_trans_stmt) is pyvex.stmt.WrTmp
            addexpr = imaginary_trans_stmt.data
            assert type(addexpr) is pyvex.expr.Binop
            assert addexpr.op == "Iop_Add27"
            arg1, arg2 = addexpr.args
            assert type(arg1) is pyvex.expr.Const
            assert arg1.con.value == 10
            assert type(arg2) is pyvex.expr.Const
            assert arg2.con.value == 20
            return
    assert False, "Could not find matching IMark"


class Instruction_MSR(Instruction):
    bin_format = bin(0x8808F380)[2:].zfill(32)
    name = "MSR.W"

    def compute_result(self):
        a = self.constant(10, Type.int_27)
        b = self.constant(20, Type.int_27)
        a + b


class Instruction_CPSIEI(Instruction):
    bin_format = bin(0xB662)[2:].zfill(16)
    name = "CPSIE I"

    def compute_result(self):
        a = self.constant(10, Type.int_27)
        b = self.constant(20, Type.int_27)
        a + b


class Instruction_CPSIEF(Instruction):
    bin_format = bin(0xB661)[2:].zfill(16)
    name = "CPSIE F"

    def compute_result(self):
        a = self.constant(10, Type.int_27)
        b = self.constant(20, Type.int_27)
        a + b


class CortexSpotter(GymratLifter):
    instrs = [Instruction_MSR, Instruction_CPSIEI, Instruction_CPSIEF]


register(CortexSpotter, "ARMEL")


def test_tmrs():
    arch = pyvex.ARCH_ARM_LE
    ins = b"\xef\xf3\x08\x82"
    b = pyvex.block.IRSB(ins, 1, arch)
    assert b.jumpkind == "Ijk_Boring"
    assert isinstance(b.statements[1].data, pyvex.expr.Get)
    assert arch.translate_register_name(b.statements[1].data.offset) in ["sp", "r13"]
    assert isinstance(b.statements[2], pyvex.stmt.Put)


def test_tmsr():
    arch = pyvex.ARCH_ARM_LE
    inss = b"\x82\xf3\x08\x88"
    b = pyvex.block.IRSB(inss, 1, arch, opt_level=3)
    assert b.jumpkind == "Ijk_Boring"
    assert isinstance(b.statements[1].data, pyvex.expr.Get)
    assert arch.translate_register_name(b.statements[1].data.offset) == "r2"
    assert isinstance(b.statements[2], pyvex.stmt.Put)


if __name__ == "__main__":
    test_basic()
    test_embedded()
    test_tmrs()
    test_tmsr()



# pylint: disable=missing-class-docstring,no-self-use
import unittest

import pyvex


class TestCacheInvalidationOnExtend(unittest.TestCase):
    def test_cache_invalidation_on_extend(self):
        b = pyvex.block.IRSB(b"\x50", 0, pyvex.ARCH_X86)
        assert b.size == 1
        assert b.instructions == 1
        toappend = pyvex.block.IRSB(b"\x51", 0, pyvex.ARCH_X86)
        toappend.jumpkind = "Ijk_Invalid"
        toappend._direct_next = None  # Invalidate the cache because I manually changed the jumpkind
        assert not toappend.direct_next
        b.extend(toappend)
        assert b.size == 2
        assert b.instructions == 2
        assert not b.direct_next


if __name__ == "__main__":
    unittest.main()



import pyvex


def test_ud2():
    # On x86 and amd64, ud2 is a valid 2-byte instruction that means "undefined instruction". Upon decoding a basic
    # block that ends with ud2, we should treat it as an explicit NoDecode, instead of skipping the instruction and
    # resume lifting.

    b = pyvex.block.IRSB(b"\x90\x90\x0f\x0b\x90\x90", 0x20, pyvex.ARCH_AMD64)
    assert b.jumpkind == "Ijk_NoDecode"
    assert b.next.con.value == 0x22
    assert b.size == 4


if __name__ == "__main__":
    test_ud2()
